{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE OverloadedStrings #-}

module GitLab.UserSearch where

import qualified Control.Exception as Exception
import Control.Monad.IO.Class (MonadIO (liftIO))
import Control.Monad.Reader (asks)
import Data.Aeson (FromJSON, eitherDecode)
import qualified Data.ByteString.Lazy as BSL
import Data.Text (Text)
import qualified Data.Text as T
import qualified Data.Text.Encoding as T
import GitLab.Types
  ( GitLab,
    GitLabServerConfig (retries, timeout, token, url),
    GitLabState (httpManager, serverCfg),
    User,
  )
import Network.HTTP.Client
  ( HttpException,
    Manager,
    Request (requestHeaders, responseTimeout),
    Response (responseBody, responseHeaders, responseStatus),
    httpLbs,
    parseRequest_,
    responseTimeoutMicro,
  )
import Network.HTTP.Types (urlEncode)
import Network.HTTP.Types.Status (Status (Status))
import Text.Read (readMaybe)

newtype GitLabUserSearchException = GitLabUserSearchException String
  deriving (Eq, Show)

instance Exception.Exception GitLabUserSearchException

-- | Searches for a user given a Username or Email. Returns @Just User@ if the
-- user is found, otherwise @Nothing@.
searchUsers ::
  -- | User to search for (Username, Email, etc)
  Text ->
  GitLab [User]
searchUsers query = do
  let path = "/search"
      attrs = "&scope=users&search=" <> query
  res <- gitlabReqJsonMany path attrs
  case res of
    Left _ -> return []
    Right users -> return users

gitlabReqJsonMany :: (FromJSON a) => Text -> Text -> GitLab (Either Status [a])
gitlabReqJsonMany urlPath attrs =
  go 1 []
  where
    go i accum = do
      cfg <- asks serverCfg
      manager <- asks httpManager
      let url' =
            url cfg
              <> "/api/v4"
              <> urlPath
              <> "?per_page=100"
              <> "&page="
              <> T.pack (show i)
              <> T.decodeUtf8 (urlEncode False (T.encodeUtf8 attrs))
      let request' = parseRequest_ (T.unpack url')
          request =
            request'
              { requestHeaders =
                  [("PRIVATE-TOKEN", T.encodeUtf8 (token cfg))],
                responseTimeout = responseTimeoutMicro (timeout cfg)
              }
      -- here
      resp <- liftIO $ tryGitLab 0 request (retries cfg) manager Nothing
      if successStatus (responseStatus resp)
        then do
          moreResults <- liftIO $ parseBSMany (responseBody resp)
          let numPages = totalPages resp
              accum' = accum ++ moreResults
          if numPages == i
            then return (Right accum')
            else go (i + 1) accum'
        else return (Left (responseStatus resp))

tryGitLab ::
  -- | the current retry count
  Int ->
  -- | the GitLab request
  Request ->
  -- | maximum number of retries permitted
  Int ->
  -- | HTTP manager
  Manager ->
  -- | the exception to report if maximum retries met
  Maybe HttpException ->
  IO (Response BSL.ByteString)
tryGitLab i request maxRetries manager lastException
  | i == maxRetries = error (show lastException)
  | otherwise =
    httpLbs request manager
      `Exception.catch` \ex -> tryGitLab (i + 1) request maxRetries manager (Just ex)

totalPages :: Response a -> Int
totalPages resp =
  let hdrs = responseHeaders resp
   in findPages hdrs
  where
    findPages [] = 1 -- error "cannot find X-Total-Pages in header"
    findPages (("X-Total-Pages", bs) : _) =
      case readMaybe (T.unpack (T.decodeUtf8 bs)) of
        Just s -> s
        Nothing -> error "cannot find X-Total-Pages in header"
    findPages (_ : xs) = findPages xs

successStatus :: Status -> Bool
successStatus (Status n _msg) =
  n >= 200 && n <= 226

parseBSMany :: FromJSON a => BSL.ByteString -> IO [a]
parseBSMany bs =
  case eitherDecode bs of
    Left s -> Exception.throwIO $ GitLabUserSearchException s
    Right xs -> return xs
