{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE RecordWildCards #-}

module Shift.Git where

import CMarkGFM (Node (Node), NodeType (DOCUMENT))
import Control.Lens ((^.))
import Control.Monad.Reader (runReaderT)
import Control.Monad.State (runStateT)
import Control.Monad.Trans (liftIO)
import Data.ByteString.Char8 (ByteString)
import qualified Data.ByteString.Char8 as BS
import Data.Default (def)
import Data.Either (rights)
import Data.Git
  ( Commit (commitMessage),
    Git,
    Ref,
    RefName (refNameRaw),
    getObject,
    tagList,
    withRepo,
  )
import Data.Git.Named (RefName (RefName))
import Data.Git.Ref (HashAlgorithm, fromHex, isHex)
import Data.Git.Storage.Object (Object (ObjCommit))
import Data.List (find, sortBy)
import Data.Maybe (catMaybes)
import Data.Set as S (Set, toList)
import Data.String.Conversions (cs)
import Data.Text (Text)
import qualified Data.Text as T
import Data.Tuple (swap)
import Data.Versions (versioning)
import Data.Void (Void)
import GitHub.Auth (Auth (OAuth))
import GitLab (GitLabServerConfig (token, url), defaultGitLabServer)
import Network.HTTP.Client (newManager)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import Shift.CLI
import Shift.Processing (generateReport, parseCommit)
import Shift.Rendering (printReport)
import Shift.Types
import Shift.Utilities (orError, pairs)
import System.Process (readCreateProcess, shell)
import Text.Megaparsec.Error (ParseErrorBundle)

parseTag :: RefName -> Either (ParseErrorBundle Text Void) TagRef
parseTag ref = case versioning . cs . refNameRaw $ ref of
  Left e -> Left e
  Right v -> Right (TagRef ref (Just v))

renderToNode :: ShiftOptions -> IO Node
renderToNode opts = withRepo ".git" $ \repo -> do
  pairedTags <- pairedTagsFromOpts opts repo

  nodes <- case opts ^. soHostingType of
    GitHubType -> do
      state <- initGitHubState

      fst
        <$> runReaderT
          (runStateT (mapM (renderDiff repo) pairedTags) state)
          opts
    GitLabType -> do
      state <- initGitLabState

      fst
        <$> runReaderT
          (runStateT (mapM (renderDiff repo) pairedTags) state)
          opts
    GitType ->
      fst
        <$> runReaderT
          (runStateT (mapM (renderDiff repo) pairedTags) GitClientState)
          opts

  pure (Node Nothing DOCUMENT (concat nodes))
  where
    initGitHubState = do
      manager <- newManager tlsManagerSettings

      token <- (opts ^. soGitHubToken) `orError` SEMissingGitHubToken
      repositoryOwner <- (opts ^. soGitHubOwner) `orError` SEMissingGitHubOwner
      repositoryName <-
        (opts ^. soGitHubRepository)
          `orError` SEMissingGitHubRepository

      pure
        GitHubClientState
          { _gcsCache = def,
            _gcsAuth = OAuth (cs token),
            _gcsManager = manager,
            _gcsOwner = cs repositoryOwner,
            _gcsRepository = cs repositoryName
          }
    initGitLabState = do
      manager <- newManager tlsManagerSettings

      server <- (opts ^. soGitLabServer) `orError` SEMissingGitLabServer
      token' <- (opts ^. soGitLabToken) `orError` SEMissingGitLabToken
      repositoryOwner <- (opts ^. soGitLabOwner) `orError` SEMissingGitLabOwner
      repositoryName <-
        (opts ^. soGitLabRepository)
          `orError` SEMissingGitLabRepository

      pure
        GitLabClientState
          { _glcsCache = def,
            _glcsServerCfg =
              ( defaultGitLabServer
                  { url = cs server,
                    token = cs token'
                  }
              ),
            _glcsHttpManager = manager,
            _glcsOwner = cs repositoryOwner,
            _glcsRepository = cs repositoryName
          }

pairedTagsFromOpts :: HashAlgorithm hash => ShiftOptions -> Git hash -> IO [(TagRef, TagRef)]
pairedTagsFromOpts opts repo = case opts of
  ShiftOptions {_soFromRef = Just fromRefRaw, _soToRef = Just toRefRaw, ..} -> do
    pure [(TagRef (RefName fromRefRaw) Nothing, TagRef (RefName toRefRaw) Nothing)]
  ShiftOptions {_soFromRef = Just fromRefRaw, _soIncludeHead, ..} -> do
    tags <- tagList repo

    let pairs = refsPairsFromRefs $ maybeIncludeHead _soIncludeHead $ parseAndSortRefs tags
        fromRef = RefName fromRefRaw

    pure $ case find (\pair -> (_tRef $ fst pair) == fromRef) pairs of
      Just matchedPair -> [matchedPair]
      Nothing -> []
  ShiftOptions {_soToRef = Just toRefRaw, ..} -> do
    tags <- tagList repo

    let pairs = refsPairsFromRefs $ maybeIncludeHead _soIncludeHead $ parseAndSortRefs tags
        toRef = RefName toRefRaw

    pure $ case find (\pair -> (_tRef $ snd pair) == toRef) pairs of
      Just matchedPair -> [matchedPair]
      Nothing -> []
  ShiftOptions {_soIncludeHead, ..} -> do
    tags <- tagList repo

    pure $ refsPairsFromRefs $ maybeIncludeHead _soIncludeHead $ parseAndSortRefs tags
  where
    maybeIncludeHead includeHead tagRefs = if includeHead then [TagRef "HEAD" Nothing] <> tagRefs else tagRefs

-- Takes a list of `TagRef`s and builds a list of tuples (e.g. From/To tuples)
refsPairsFromRefs :: [TagRef] -> [(TagRef, TagRef)]
refsPairsFromRefs refs = swap <$> pairs refs

-- Takes a set of `RefName`s and returns a sorted list of `TagRef`s, ordered in
-- descending version numbers.
parseAndSortRefs :: Set RefName -> [TagRef]
parseAndSortRefs tags = sortBy (flip compare) . rights $ parseTag <$> toList tags

renderDiff ::
  (ClientState s, HashAlgorithm hash) =>
  Git hash ->
  (TagRef, TagRef) ->
  GitM s [Node]
renderDiff repo (tx, ty) = do
  diff <- lookupCommitsDiff repo tx ty

  case diff of
    [] -> printReport (tx, ty) def
    diff_ -> printReport (tx, ty) (generateReport . rights $ parseCommit <$> diff_)

lookupCommitsDiff ::
  (HashAlgorithm hash) =>
  Git hash ->
  TagRef ->
  TagRef ->
  GitM s [(Ref hash, Commit hash)]
lookupCommitsDiff repo x y = do
  rawOutput <- liftIO $ readCreateProcess (shell gitCommand) ""

  catMaybes
    <$> mapM
      (lookupRawRef repo)
      (filter (not . T.null) . T.splitOn "\n" . cs $ rawOutput)
  where
    gitCommand =
      mconcat
        [ "git rev-list ",
          refNameRaw . _tRef $ x,
          "...",
          refNameRaw . _tRef $ y
        ]

lookupRawRef ::
  (HashAlgorithm hash) =>
  Git hash ->
  Text ->
  GitM s (Maybe (Ref hash, Commit hash))
lookupRawRef repo rr =
  if isHex . cs $ rr
    then do
      let ref = fromHex . cs $ rr

      object <- liftIO $ getObject repo ref True

      case object of
        Just (ObjCommit commit) -> pure $ Just (ref, commit)
        _ -> pure Nothing
    else pure Nothing

commitSummary :: Commit hash -> ByteString
commitSummary = head . BS.split '\n' . commitMessage
