{-# LANGUAGE ApplicativeDo #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE ViewPatterns #-}

module Shift.Parsers where

import Control.Applicative (optional)
import Control.Monad (void)
import Data.HashSet (HashSet, fromList)
import Data.List.NonEmpty (NonEmpty ((:|)))
import qualified Data.List.NonEmpty as NE
import Data.String.Conversions (cs)
import Data.Text (Text)
import Data.Void (Void)
import Shift.Types
  ( BreakingChange (BreakingChange),
    CommitType (..),
    ConventionalCommit (ConventionalCommit),
    MergeCommit (..),
    MiscCommit (..),
    ParsedCommit (..),
    TicketChange,
  )
import Text.Megaparsec
  ( MonadParsec (eof, lookAhead, try),
    Parsec,
    Stream (Token),
    anySingle,
    choice,
    many,
    manyTill,
    skipSome,
    some,
    someTill,
    (<|>),
  )
import Text.Megaparsec.Char (char, eol, spaceChar, string)

type Parser = Parsec Void Text

commit :: Parser ParsedCommit
commit =
  (PCConventional <$> conventionalCommit)
    <|> try (PCMerge <$> mergeCommit)
    <|> (PCMisc <$> miscCommit)

spaced :: Parser a -> Parser a
spaced = (*>) $ many (void (some spaceChar) <|> void eol)

oneParserOf :: NE.NonEmpty (Parser a) -> Parser a
oneParserOf (NE.reverse -> (x :| xs)) = foldl (\acc y -> try y <|> acc) x xs

manyCharsTill :: Parser end -> Parser [Token Text]
manyCharsTill = manyTill anySingle

someCharsTill :: Parser end -> Parser [Token Text]
someCharsTill = someTill anySingle

conventionalCommit :: Parser ConventionalCommit
conventionalCommit = do
  cType <- commitType
  cScope <- optional (char '(' *> someCharsTill (try $ char ')'))
  void $ string ": "
  cSubject <- someCharsTill (choice [void eol, eof])

  cBody <- spaced commitBody

  cBreakingChanges <- many (try breakingChange)
  cTicketChanges <- many (try ticketChange)

  pure $
    ConventionalCommit
      cType
      (cs <$> cScope)
      (cs cSubject)
      (cs cBody)
      cBreakingChanges
      cTicketChanges

commitType :: Parser CommitType
commitType = do
  cType <-
    oneParserOf . fmap string $
      "feat"
        :| [ "fix",
             "bug",
             "docs",
             "doc",
             "style",
             "refactor",
             "ref",
             "test",
             "chore"
           ]

  pure $ case cType of
    "feat" -> CTFeature
    "fix" -> CTFix
    "bug" -> CTFix
    "docs" -> CTDocs
    "doc" -> CTDocs
    "style" -> CTStyle
    "refactor" -> CTRefactor
    "ref" -> CTRefactor
    "test" -> CTTest
    "chore" -> CTChore
    _ -> CTFeature

commitBody :: Parser String
commitBody = manyCharsTill sectionEnd
  where
    sectionEnd = try (spaced eof) <|> lookAheadToBreakingChange <|> lookAheadToTicketChange
    lookAheadToBreakingChange = try . void $ lookAhead (spaced breakingChange)
    lookAheadToTicketChange = try . void $ lookAhead (spaced ticketChange)

breakingChange :: Parser BreakingChange
breakingChange = do
  void . spaced . string $ "BREAKING CHANGE: "

  do
    shortDescription <- cs <$> manyCharsTill eol
    body <- optional (cs <$> (void eol *> someCharsTill sectionEnd))

    return $ BreakingChange shortDescription body
  where
    sectionEnd = try (spaced eof) <|> lookAheadToTicketChange
    lookAheadToTicketChange = try . void $ lookAhead (eol *> eol *> spaced ticketChange)

ticketChange :: Parser (HashSet TicketChange)
ticketChange = do
  tcAction <- spaced $ manyCharsTill (string ": ")
  tcTickets <-
    some
      ( char '#'
          *> manyCharsTill
            (void (string ", ") <|> void (some spaceChar) <|> void eol)
      )

  pure . fromList $ (\x -> (cs tcAction, cs x)) <$> tcTickets

mergeCommit :: Parser MergeCommit
mergeCommit =
  MergeCommit . cs
    <$> (string "Merge " *> manyCharsTill (skipSome eol <|> eof))

miscCommit :: Parser MiscCommit
miscCommit = MiscCommit . cs <$> manyCharsTill (skipSome eol <|> eof)
