{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE OverloadedStrings #-}

module Shift.Rendering where

import CMarkGFM as M
import Control.Lens ((^.))
import Control.Monad.Catch (MonadThrow)
import Control.Monad.IO.Class (MonadIO)
import Control.Monad.State (MonadState)
import Control.Monad.Trans.Writer (execWriterT)
import Control.Monad.Trans.Writer.Lazy (tell)
import Data.Default (Default (def))
import Data.Git (Commit, Ref, RefName (refNameRaw), commitAuthor)
import Data.List (sortOn)
import Data.Maybe (catMaybes, fromMaybe)
import Data.String.Conversions (cs)
import Data.Text (Text)
import qualified Data.Text as T (pack, take)
import Data.Versions (prettyV)
import Shift.Types
  ( ChangeReport,
    ClientState (..),
    ConventionalGroup,
    MergeCommit (MergeCommit),
    MergeGroup,
    MiscCommit (MiscCommit),
    MiscGroup,
    TagRef (_tRef, _tVersioning),
    ccScope,
    ccSubject,
    crChores,
    crDocs,
    crFeatures,
    crFixes,
    crMerges,
    crMisc,
    crRefactors,
    crStyles,
    crTests,
  )

-- $setup
-- >>> :set -XOverloadedStrings

renderRef ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  Ref hash ->
  m [Node]
renderRef ref = do
  url <- getRefURL ref

  let shortRef = T.take 7 . cs . show $ ref
  let shortRefNode = leafNode (CODE shortRef)

  pure $ case url of
    Just url_ -> [textNode "[", parentNode (LINK url_ shortRef) [shortRefNode], textNode "]"]
    Nothing -> [textNode "[", shortRefNode, textNode "]"]

renderConventionalCommit ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  ConventionalGroup hash ->
  m Node
renderConventionalCommit (ref, commit, pc) = do
  renderedRef <- renderRef ref
  authorText <- renderAuthor commit ref

  pure $
    listItemNode
      [ paragraphNode $
          renderedRef
            <> catMaybes
              [ Just $ textNode " ",
                Just . bold $ fromMaybe "" (pc ^. ccScope) <> ":",
                Just $ textNode " ",
                Just . textNode $ pc ^. ccSubject,
                Just $ textNode " ",
                authorText
              ]
      ]

renderAuthor ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  Commit hash ->
  Ref hash ->
  m (Maybe Node)
renderAuthor commit ref = do
  authorInfo <- getAuthorInfo (commitAuthor commit) ref

  pure $ case authorInfo of
    Just (username, authorUrl) -> Just $ parentNode (LINK authorUrl username) [textNode username]
    Nothing -> Nothing

renderMiscCommit ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  MiscGroup hash ->
  m Node
renderMiscCommit (ref, commit, MiscCommit subject) = do
  renderedRef <- renderRef ref
  authorText <- renderAuthor commit ref

  pure $
    listItemNode
      [ paragraphNode $
          renderedRef
            <> catMaybes
              [ Just $ textNode subject,
                Just $ textNode " ",
                authorText
              ]
      ]

renderMergeCommit ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  MergeGroup hash ->
  m Node
renderMergeCommit (ref, commit, MergeCommit subject) = do
  renderedRef <- renderRef ref
  authorText <- renderAuthor commit ref

  pure $
    listItemNode
      [ paragraphNode $
          renderedRef
            <> catMaybes
              [ Just $ textNode $ "Merge " <> subject,
                authorText
              ]
      ]

-- | Renders a range of `TagRef`s.
--
-- Examples:
--
-- >>> import Shift.Types (TagRef(TagRef))
-- >>> renderRange (TagRef "1.0.0" Nothing) (TagRef "1.1.0" Nothing)
-- "1.0.0 to 1.1.0"
renderRange :: TagRef -> TagRef -> Text
renderRange tx ty =
  mconcat
    [renderTagRef tx, " to ", renderTagRef ty]

-- | Renders a `TagRef`.
--
-- >>> import Data.Versions
-- >>> import Shift.Types (TagRef(TagRef))
-- >>> let version = Ideal $ SemVer 1 0 0 [] []
-- >>> renderTagRef (TagRef "1.0.0" (Just version))
-- "1.0.0"
--
-- >>> import Data.Versions
-- >>> import Shift.Types (TagRef(TagRef))
-- >>> let version = Ideal $ SemVer 1 0 0 [] []
-- >>> renderTagRef (TagRef "1.0.1" (Just version))
-- "1.0.0"
--
-- >>> import Shift.Types (TagRef(TagRef))
-- >>> renderTagRef (TagRef "1.0.1" (Nothing))
-- "1.0.1"
renderTagRef :: TagRef -> Text
renderTagRef tr = case _tVersioning tr of
  Just versioning -> prettyV versioning
  Nothing -> T.pack . refNameRaw $ _tRef tr

bold :: Text -> Node
bold x = parentNode STRONG [textNode x]

indented :: Int -> Text -> Text
indented levels x = foldl (\acc _ -> acc <> "  ") "" [1 .. levels] <> x

indentedL :: Int -> Text -> Text
indentedL levels x = "\\" <> foldl (\acc _ -> acc <> "-") "" [1 .. levels] <> x

headerOne :: Text -> Node
headerOne t = parentNode (HEADING 1) [textNode t]

headerTwo :: Text -> Node
headerTwo t = parentNode (HEADING 2) [textNode t]

headerThree :: Text -> Node
headerThree t = parentNode (HEADING 3) [textNode t]

textNode :: Text -> Node
textNode t = Node Nothing (TEXT t) []

leafNode :: NodeType -> Node
leafNode n = Node Nothing n []

parentNode :: NodeType -> [Node] -> Node
parentNode = Node Nothing

listItemNode :: [Node] -> Node
listItemNode = parentNode ITEM

paragraphNode :: [Node] -> Node
paragraphNode = parentNode PARAGRAPH

printReport ::
  (MonadIO m, MonadState s m, ClientState s, MonadThrow m) =>
  (TagRef, TagRef) ->
  ChangeReport hash ->
  m [Node]
printReport (rangeStart, rangeEnd) report = execWriterT $ do
  tell [headerOne $ renderRange rangeStart rangeEnd]

  -- If the report is empty, don't render any changes.
  if report == def
    then tell [paragraphNode [textNode "No changes"]]
    else do
      -- Print conventional commits
      conventionalSection "New features:" crFeatures
      conventionalSection "Fixes:" crFixes
      conventionalSection "Updated documentation:" crDocs
      conventionalSection "Refactors:" crRefactors
      conventionalSection "Updated tests:" crTests
      conventionalSection "Style fixes:" crStyles
      conventionalSection "Chores:" crChores

      -- Print miscs and merges
      miscCommits <- mapM renderMiscCommit (report ^. crMisc)
      mergeCommits <- mapM renderMergeCommit (report ^. crMerges)

      textSection "Miscellaneous changes:" miscCommits
      textSection "Merge commits:" mergeCommits
  where
    conventionalSection label sectionLens = do
      let rawCommits = sortOn (\(_, _, pc) -> pc ^. ccScope) (report ^. sectionLens)

      commits <- mapM renderConventionalCommit rawCommits
      textSection label commits

    textSection label commits = case commits of
      [] -> pure ()
      _ -> do
        tell [headerTwo label]

        tell [parentNode (LIST (ListAttributes BULLET_LIST False 0 PAREN_DELIM)) commits]
